#!/usr/bin/env python

__author__ = "branko@toic.org (http://toic.org)"
__date__ = "Dec 24, 2012 0:10 PM$"
__version__ = "0.3.0b"


from ApTop import ApacheStatus
from ApTop import AptopCurses


if __name__ == "__main__":
    aptop = ApacheStatus()
    window = AptopCurses(aptop)



__author__ = "branko@toic.org (http://toic.org)"
__date__ = "Dec 24, 2012 0:10 PM$"
__version__ = "0.3.0b"

import ConfigParser
import os
import sys
from operator import itemgetter

try:
    import lxml.html  # @UnresolvedImport
except:
    print "No lxml package found..."
    print "please install lxml >= 3.x.x"
    print "best way to install easy_install lxml"
    sys.exit(1)


class ApacheStatus(object):
    def __init__(self):
        """
        Try to detect config file location, trying to use user defined config
        in home path ~/.aptop.conf or /etc/aptop.conf then
        """
        homedir = os.path.expanduser('~')

        if os.path.isfile(os.path.join(homedir, '.aptop.conf')):
            self.configfile = os.path.join(homedir, '.aptop.conf')
        else:
            self.configfile = None

        """
        Let's populate some defaults if no config file is found
        """
        if not self.configfile:
            if os.path.isdir('/var/cpanel'):
                self.status_url = 'http://localhost/whm-server-status'
            else:
                self.status_url = 'http://localhost/server-status'
            self.refresh = '5'
        else:
            config = ConfigParser.ConfigParser()
            config.readfp(open(self.configfile))
            try:
                self.status_url = config.get('aptop', 'status_url')
            except:
                if os.path.isdir('/var/cpanel'):
                    self.status_url = 'http://localhost/whm-server-status'
                else:
                    self.status_url = 'http://localhost/server-status'
            try:
                self.refresh = config.get('aptop', 'refresh')
            except:
                self.refresh = '5'
        try:
            self.tree = lxml.html.parse(self.status_url)
        except:
            print "Apache not running or wrong mod_status url!?"
            sys.exit(1)

        # define should we filter out inactive sessions
        self.active = True
        self.sort_by = 'SS'
        # key case must be defined exactly as named in data source
        self.sort_fields = {'SS': 'float',
                            'CPU': 'float',
                            'Req': 'float',
                            'Conn': 'float',
                            'VHost': 'str',
                            'Request': 'str'
                            }
        # this is passed to reverse parameter on sort(list)
        self.sort_order = False

    def sort_options(self):
        """ (NoneType) -> list of string and dict

           Returns the current self.sort_fields and self.sort_by variable
        """
        return [self.sort_by, self.sort_fields]

    def reverse_order(self):
        """ just reverese the ordering of current sorting """

        if self.sort_order:
            self.sort_order = False
        else:
            self.sort_order = True

    def refresh_rate(self):
        """
        (NoneType) -> int

        Returns parsed refresh time interval as int
        """
        return int(self.refresh)

    def fetch_status(self):
        """
        (NoneType) -> NoneType

        Refetching of data
        """
        if self.tree:
            old_tree = self.tree
        try:

            self.tree = lxml.html.parse(self.status_url)
        except:
            self.tree = old_tree

    def verify_mod_status(self):
        """
        (str) -> boolean

        Checks the data string for Apache Status title and returns true or
        false
        """
        return self.tree.find('.//title').text == 'Apache Status'

    def count_by_vhost(self, data):
        """
        (str) -> list of tuple

        Counts the active concurent connections by vhosts and returns an
        ordered list of tuples containing vhost name and number of active
        connections
        """
        vstatus = {}
        if self.active:
            vhosts = self.filter_active(data)
        else:
            vhosts = data
        for status in vhosts:
            if status['VHost'] in vstatus:
                vstatus[status['VHost']] += 1
            else:
                vstatus[status['VHost']] = 1
        items = vstatus.items()
        items.sort(key=itemgetter(1), reverse=self.sort_order)

        return items

    def count_by_client(self, data):
        """
        (str) -> list of tuple

        Counts the active concurent connections by clients and returns an
        ordered list of tuples containing client IP and number of active
        connections
        """
        cstatus = {}
        if self.active:
            vhosts = self.filter_active(data)
        else:
            vhosts = data
        for status in vhosts:
            if status['Client'] in cstatus:
                cstatus[status['Client']] += 1
            else:
                cstatus[status['Client']] = 1
        items = cstatus.items()
        items.sort(key=itemgetter(1), reverse=self.sort_order)

        return items

    def count_and_group_requests_by_vhost(self, data):
        """
        (str) -> list of tuples

        Groups, counts and sorts the requests per vhost.
        Then sorts the resulting list by the total number
        of requests per vhost
        """
        grouped = {}
        if self.active:
            vhosts = self.filter_active(data)
        else:
            vhosts = data

        for status in vhosts:

            vhost = status['VHost']
            req_uri = status['Request']

            if vhost in grouped:
                if req_uri in grouped[vhost]:
                    grouped[vhost][req_uri] += 1
                else:
                    grouped[vhost][req_uri] = 1
            else:
                grouped[vhost] = {}
                grouped[vhost][req_uri] = 1

        # TODO: not loving this, but it seems to work for now
        for vhost in grouped:
            vhost_requests = sorted(
                grouped[vhost].items(),
                key=lambda x: (x[1], x[0]),
                reverse=self.sort_order
            )
            count = sum(x[1] for x in vhost_requests)
            grouped[vhost] = {'reqs': vhost_requests, 'cnt': count}

        grouped_sorted = sorted(
            grouped.items(),
            key=lambda x: (x[1].values()[0], x[0]),
            reverse=self.sort_order
        )
        return grouped_sorted

    def count_by_request(self, data):
        """
        (str) -> list of tuples

        Counts the active requests and returns an ordered
        list of tuples containing the request URI and the matching number
        of active connections
        """
        rstatus = {}
        if self.active:
            vhosts = self.filter_active(data)
        else:
            vhosts = data
        for status in vhosts:
            if status['Request'] in rstatus:
                rstatus[status['Request']] += 1
            else:
                rstatus[status['Request']] = 1
        items = rstatus.items()
        items.sort(key=itemgetter(1), reverse=self.sort_order)

        return items

    def togle_active(self):
        if self.active:
            self.active = False
        else:
            self.active = True

    def update_sort_field(self, field):
        for key in self.sort_fields:
            if field.lower() == key.lower():
                self.sort_by = key
                return True
        return False

    def filter_active(self, data):
        """
        (list of dict) -> list of dict

        Returns list of dicts based on filter status

        """
        filtered = []

        for status in data:
            if status['M'] != '_' and status['M'] != '.':
                filtered.append(status)
        return filtered

    def display_vhosts(self, data):
        """
        (list of dict) -> list of dict
        retrurn vhost data as list of dicts based on current filter
        """
        if self.active:
            results = self.filter_active(data)
        else:
            results = data
        return results

    def sort_vhosts_by(self, values, sort_method):

        if sort_method == 'float':
            return sorted(
                values,
                key=lambda k: float(k[self.sort_by]),
                reverse=self.sort_order
            )
        elif sort_method == 'str':
            return sorted(
                values,
                key=lambda k: str(k[self.sort_by]),
                reverse=self.sort_order
            )

    def parse_vhosts(self):
        """
        (str) -> list of dict

        Parses a apache status from internal class self.tree variable
        and returns list ofdict containing all vhost informations from
        mod_status

        """
        tree = self.tree.xpath('//table[@border="0"]')[0]
        vhost_status = []
        headers = tree.findall('.//th')
        h2 = [s.text_content().replace('\n', '') for s in headers]
        h2.append("Method")
        for row in tree.findall('.//tr')[1:]:  # this is header, excluding
            d = [
                s.text_content().replace('\n', '')
                for s in row.findall('.//td')
            ]

            d.append(d[-1].split()[0])
            vhost_status.append(dict(zip(h2, d)))

        return self.sort_vhosts_by(
            vhost_status,
            self.sort_fields[self.sort_by]
        )

    def parse_header(self):
        """
        (NoneType) -> list

        Returns a dictionary of header status variables from apache status
        page.

        Available header fields:
           -  Server Version
           -  Server Built
           -  Current Time
           -  Restart Time
           -  Server uptime
           -  Parent Server Generation
           -  CPU Usage
           -  Total accesses
           -  Total Traffic
           -  working childs #requests currently being processed by the server
           -  idle childs    #idle workers
           -  requests/sec
           -  B/second
           -  kB/request

        """

        HEADER_LIST = [
            'Server Version:',
            'Server Built:',
            'Current Time:',
            'Restart Time:',
            'Parent Server Generation:',
            'Server uptime:',
            'Total accesses:',
            'CPU Usage:',
            'requests',
            'workers',
        ]

        headers = {}

        for h in self.tree.findall('.//dt'):
            line = h.text.replace('\n', '')
            for item in HEADER_LIST:
                if item in line:

                    if item == 'workers':
                        for req in line.split(','):
                            req = req.strip()
                            if req.split()[-1] == 'processed':
                                headers['working childs'] = req.split()[0]
                            elif req.split()[-1] == 'workers':
                                headers['idle childs'] = req.split()[0]
                    elif item == 'requests':
                        for req in line.split('-'):
                            req = req.strip()
                            headers[req.split()[1]] = req.split()[0]

                    elif item == 'Total accesses:':
                        for el in line.split('-'):
                            el = el.strip()
                            key = el.split(':')[0].strip()
                            value = el.split(':')[1].strip()
                            headers[key] = value

                    else:
                        headers[item[:-1]] = line.split(':')[1].strip()

        return headers



import sys
assert sys.version >= '2' , 'Install Python 2.0 or greater'

from setuptools import setup, find_packages

# $Id: setup.py,v 1.2 2007/05/10 16:17:05 jay Exp $

def runSetup ():
    setup(name='aptop',
          version='0.3.0b',
          author='Branko Toic' ,
          author_email='branko@toic.org' ,
          url='https://bitbucket.org/btoic/aptop/get/master.tar.gz' ,
          license='GPLv2' ,
          install_requires=['lxml'],
          setup_requires=['lxml'],
          platforms=[ 'unix' ] ,
          description='ApTop is top like clone for apaache geting info from' \
              ' apache mod_status' ,
          packages=find_packages(),
          scripts=['aptop.py'],
          provides=['ApTop'],
    )

if __name__ == '__main__':
    runSetup()



__author__ = "branko@toic.org (http://toic.org)"
__date__ = "Dec 24, 2012 0:10 PM$"
__version__ = "0.3.0b"

import curses
from _curses import error as CursesError
import sys
import os
HEADER_HEIGHT = 5
FOOTER_HEIGHT = 2


class AptopCurses(object):
    def __init__(self, aptop):

        self.stdscr = curses.initscr()
        curses.noecho()
        curses.raw()
        curses.nonl()
        # Some terminals don't support different cursor visibilities
        try:
            curses.curs_set(0)
        except CursesError:
            pass

        # Some terminals don't support the default color palette
        try:
            curses.start_color()
            curses.use_default_colors()
        except CursesError:
            pass
        self.aptop = aptop
        self.MAX_H, self.MAX_W = self.stdscr.getmaxyx()
        self.running = True
        self.draw_view_keys = {
            'V': self.draw_vhosts,
            'H': self.draw_dashboard,
            'C': self.draw_clients,
            'U': self.draw_vhosts_with_uris
        }

        self.handle_view_keys = {
            'I': self.aptop.togle_active,
            'R': self.aptop.reverse_order,
            'Q': self.aptop_stop,
            'D': self.draw_update_refresh,
            'O': self.draw_update_order,
        }

        self.stdscr.nodelay(1)
        self.refresh = self.aptop.refresh_rate()
        self.counter = 0
        self.BODY_H = self.MAX_H
        self.view = 'H'
        curses.wrapper(self.start)

    def start(self, stdscr):
        if self.aptop.verify_mod_status():
            while self.running:
                self.aptop.fetch_status()
                self.BODY_H = self.MAX_H - HEADER_HEIGHT - FOOTER_HEIGHT
                c = self.stdscr.getch()
                if c in range(255):
                    key = str.upper(chr(c))
                    if key in self.draw_view_keys:
                        self.view = key
                    elif key in self.handle_view_keys:
                        self.handle_view_keys[key]()
                    self.counter = 0
                self.iterate()
        else:
            print "Apache not running or wrong mod_status url!"
            sys.exit(1)

    def draw_view(self):
        self.draw_header()
        # already filtered in the main loop
        self.draw_view_keys[self.view]()
        self.draw_footer()

    def iterate(self):
        if self.counter == 0:
            self.draw_view()
            self.counter = self.refresh * 2
        else:
            self.counter -= 1

        curses.napms(500)
        self.MAX_H, self.MAX_W = self.stdscr.getmaxyx()

    def aptop_stop(self):
        """ calling this will halt aptop"""
        if self.running:
            self.running = False

    """ Drawing output """

    def draw_update_refresh(self):
        ref = curses.newwin(self.MAX_H, self.MAX_W, 0, 0)

        running = True

        ref.addstr(10, 10, str('New refresh time in sec?'))
        # input handling
        acumulate = ''
        counter = 0
        while running:

            counter += 1
            c = self.stdscr.getch()
            if c in [curses.KEY_ENTER, curses.KEY_BREAK, 13]:
                try:
                    refresh = int(acumulate)
                    self.refresh = refresh
                    running = False
                except:
                    ref.addstr(12, 40, str('Not a valid input'))
                    ref.addstr(10, 35, str(' ') * (len(acumulate) + 20))
                    ref.addstr(10, 34, str(' '))
                    acumulate = ''
            else:
                if c in range(255):
                    if len(acumulate) < 1:
                        ref.addstr(12, 40, str('                  '))
                    try:
                        acumulate += chr(c)
                        ref.addstr(10, 35, str(acumulate))
                    except:
                        pass
            ref.refresh()
        # close input and exit the loop

    def draw_update_order(self):
        order = curses.newwin(self.MAX_H, self.MAX_W, 0, 0)

        running = True
        current = self.aptop.sort_options()
        order.addstr(1, 10, 'Available options are')
        opcount = 2
        for option in current[1]:
            if current[0] == option:
                order.addstr(opcount, 5, '*')
                order.addstr(opcount, 10, str(option))
            else:
                order.addstr(opcount, 10, str(option))
            opcount += 1

        order.addstr(10, 10, str('Type new ordering field?'))
        # input handling
        acumulate = ''
        while running:
            c = self.stdscr.getch()
            if c in [curses.KEY_ENTER, curses.KEY_BREAK, 13]:
                if self.aptop.update_sort_field(acumulate):
                    running = False
                else:
                    order.addstr(12, 40, str('Not a valid input'))
                    order.addstr(10, 35, str(' ') * (len(acumulate) + 20))
                    order.addstr(10, 34, str(' '))
                    acumulate = ''
            else:
                if c in range(255):
                    if len(acumulate) < 1:
                        order.addstr(12, 40, str('                  '))
                    try:
                        acumulate += chr(c)
                        order.addstr(10, 35, str(acumulate))
                    except:
                        pass
            order.refresh()
        # close input and exit the loop

    def draw_header(self):
        """ draws a header window """

        header = curses.newwin(HEADER_HEIGHT, self.MAX_W, 0, 0)
        header_data = self.aptop.parse_header()
        header1 = "%-5s: %s %1s: %s" % ('System load',
                                        os.getloadavg()[0],
                                        'CPU Usage',
                                        header_data['CPU Usage'],
                                        )
        header2 = "%-5s: %s %1s: %s %1s: %s" % (
            'Total Traffic',
            header_data['Total Traffic'],
            'Total accesses',
            header_data['Total accesses'],
            'kB/request',
            header_data['kB/request'],
        )

        header3 = "%-5s: %s %1s: %s %1s: %s %1s: %s" % (
            'reqests/sec',
            header_data['requests/sec'],
            'busy childs',
            header_data['working childs'],
            'idle childs',
            header_data['idle childs'],
            'B/second',
            header_data['B/second'],
        )

        header4 = "%-5s: %s" % (
            'Server uptime',
            header_data['Server uptime'],
        )

        header.addstr(0, 1, str(header1))
        header.addstr(1, 1, str(header2))
        header.addstr(2, 1, str(header3))
        header.addstr(3, 1, str(header4))
        header.refresh()

    def draw_dashboard(self):
        """ draws a dashboard window """
        dash = curses.newwin(self.BODY_H, self.MAX_W, HEADER_HEIGHT, 0)
        dash_data = self.aptop.display_vhosts(self.aptop.parse_vhosts())
        dcount = 0
        try:
            formatstr = '%-5s %1s %7s %3s %6s %7s %12s %-15s %-25s %-50s' % (
                'PID', 'M', 'CPU', 'SS', 'Req', 'Conn', 'Acc', 'Client',
                'VHost', 'Request' + ' ' * self.MAX_W)

            dash.addstr(0, 0, formatstr, curses.A_REVERSE)
        except curses.error:
            pass
        for dash_line in dash_data:
            dcount += 1
            #dirty cpu missing fix for bug #5
            if 'CPU' in dash_line:
                cpu_value = dash_line['CPU']
            else:
                cpu_value = 'NaN'
            #end
            try:
                formatstr = '%-5s %1s %7s %3s %6s %7s %12s %-15s %-25s %s' % \
                    (
                        dash_line['PID'],
                        dash_line['M'],
                        cpu_value,
                        dash_line['SS'],
                        dash_line['Req'],
                        dash_line['Conn'],
                        dash_line['Acc'],
                        dash_line['Client'],
                        dash_line['VHost'][:26],
                        dash_line['Request']
                    )
                dash.addstr(dcount, 0, formatstr)
            except curses.error:
                pass
        dash.refresh()

    def draw_vhosts(self):
        """ draws a vhosts window """
        body = curses.newwin(self.BODY_H, self.MAX_W, HEADER_HEIGHT, 0)

        body_data = self.aptop.count_by_vhost(self.aptop.parse_vhosts())
        bcount = 0
        try:
            body.addstr(0, 0, str(" ") * self.MAX_W, curses.A_REVERSE)
            body.addstr(bcount, 1, 'Num req', curses.A_REVERSE)
            body.addstr(bcount, 20, "Virtualhost", curses.A_REVERSE)
        except curses.error:
            pass
        for body_line in body_data:
            bcount += 1
            try:
                body.addstr(bcount, 1, str(body_line[1]))
                body.addstr(bcount, 20, str(body_line[0]))
            except curses.error:
                pass
        body.refresh()

    def draw_clients(self):
        """ draws a vhosts window """
        body = curses.newwin(self.BODY_H, self.MAX_W, HEADER_HEIGHT, 0)
        body_data = self.aptop.count_by_client(self.aptop.parse_vhosts())
        bcount = 0
        try:
            body.addstr(0, 0, str(" ") * self.MAX_W, curses.A_REVERSE)
            body.addstr(bcount, 1, 'Num req', curses.A_REVERSE)
            body.addstr(bcount, 20, "Client IP", curses.A_REVERSE)
        except curses.error:
            pass
        for body_line in body_data:
            bcount += 1
            try:
                body.addstr(bcount, 1, str(body_line[1]))
                body.addstr(bcount, 20, str(body_line[0]))
            except curses.error:
                pass
        body.refresh()

    def draw_vhosts_with_uris(self):
        """ draws the 'requests-grouped-by-vhosts' window """
        body = curses.newwin(self.BODY_H, self.MAX_W, HEADER_HEIGHT, 0)

        body_data = self.aptop.count_and_group_requests_by_vhost(
            self.aptop.parse_vhosts()
        )
        linecount = 0

        try:
            body.addstr(0, 0, str(" ") * self.MAX_W, curses.A_REVERSE)
            body.addstr(linecount, 1, 'Num req', curses.A_REVERSE)
            body.addstr(linecount, 20, 'URI', curses.A_REVERSE)
        except curses.error:
            pass

        for vhost, vhost_data in body_data:
            linecount += 1
            requests = vhost_data['reqs']
            vhost_str = str(
                '>>> ' + vhost + ' (' + str(vhost_data['cnt']) + ')'
            )

            try:
                body.addstr(linecount, 1, vhost_str, curses.A_BOLD)
                linecount += 1
                for k, v in requests:
                    body.addstr(linecount, 1, str(v))
                    body.addstr(linecount, 20, str(k))
                    linecount += 1
            except curses.error:
                pass

        body.refresh()

    def draw_footer(self):
        """ draws a footer window """
        footer = curses.newwin(
            FOOTER_HEIGHT,
            self.MAX_W,
            self.MAX_H - FOOTER_HEIGHT,
            0
        )
        try:
            footer.addstr(1, 1, 'Q')
            footer.addstr(1, 3, 'Quit ApTOP', curses.A_REVERSE)
            footer.addstr(1, 14, 'H')
            footer.addstr(1, 16, 'Home', curses.A_REVERSE)
            footer.addstr(1, 26, 'V')
            footer.addstr(1, 28, 'By Vhost', curses.A_REVERSE)
            footer.addstr(1, 37, 'C')
            footer.addstr(1, 39, 'By Client', curses.A_REVERSE)
            footer.addstr(1, 50, 'U')
            footer.addstr(1, 52, 'By Vhost/URIs', curses.A_REVERSE)
        except curses.error:
            pass
        footer.refresh()
